/*-----------------------------------------------------------------------

  File  : labsolve.c

  Author: Stephan Schulz (schulz@eprover.org)

  Contents

  Program reading and solving simple ASCII labyrinths.

  Copyright 2015 by the author.
  This code is released under the GNU General Public Licence, version
  2, or, at your choice, any later version.

  Changes

  <1> Tue Jun  2 18:01:34 CEST 2015
  New
  -----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <ctype.h>
#include <limits.h>
#include <assert.h>

#define MAXCOLS  2048
#define MAXROWS  2048


/* Data type: Labyrinth - ASCII map, costs, direction */

typedef struct labyrinth
{
   char lab[MAXCOLS+2][MAXROWS];
   long costs[MAXCOLS][MAXROWS];
   int maxrow;
   int maxcol;
   int startx;
   int starty;
   int treasurex;
   int treasurey;
}LabCell, *Lab_p;


/* Data structure for represeting individual moves */

typedef struct relpos
{
   int dx;
   int dy;
}RelPos;

/* Legal moves in the base case */

RelPos straight_moves[] =
{
   { 0,  1},
   { 0, -1},
   { 1,  0},
   {-1,  0},
   {INT_MAX, INT_MAX}
};

/* Legal moves if diagonals are ok */

RelPos all_moves[] =
{
   { 0,  1},
   { 0, -1},
   { 1,  0},
   {-1,  0},
   { 1,  1},
   { 1, -1},
   {-1,  1},
   {-1, -1},
   {INT_MAX, INT_MAX}
};




/*-----------------------------------------------------------------------
//
// Function: xmalloc()
//
//   Thin wrapper around malloc() - fail noisily (with error message)
//   if no memory is available.
//
// Global Variables: -
//
// Side Effects    : Only via malloc() or in case of error
//
/----------------------------------------------------------------------*/

void* xmalloc(size_t size)
{
   void *mem = malloc(size);
   if(!mem)
   {
      fprintf(stderr, "Out of memory\n");
      exit(EXIT_FAILURE);
   }
   return mem;
}

/*-----------------------------------------------------------------------
//
// Function: xstrdup()
//
//   Thin wrapper around strdup() - fail noisily (with error message)
//   if no memory is available.
//
// Global Variables: -
//
// Side Effects    : Only via malloc() or in case of error
//
/----------------------------------------------------------------------*/

char* xstrdup(char* str)
{
   void *newstr = strdup(str);
   if(!newstr)
   {
      fprintf(stderr, "Out of memory\n");
      exit(EXIT_FAILURE);
   }
   return newstr;
}

/*-----------------------------------------------------------------------
//
// Function: LabInit()
//
//   Verify and Initialize the lab: seach for and record Start and
//   Treasure positions. End with error if they are not found.
//
// Global Variables: -
//
// Side Effects    : Only in error case.
//
/----------------------------------------------------------------------*/

void LabInit(Lab_p lab)
{
   int i,j;
   bool start = false, treasure = false;

   lab->maxcol = 0;
   for(i=0; i< lab->maxrow; i++)
   {
      for(j=0; j<MAXCOLS; j++)
      {
         if((lab->lab[i][j]=='\n') && (j>lab->maxcol))
         {
            lab->maxcol = j;
         }
         lab->costs[i][j] = LONG_MAX;
         if(lab->lab[i][j] == 'S')
         {
            if(start)
            {
               fprintf(stderr, "Error: Multiple start positions\n");
               exit(EXIT_FAILURE);
            }
            lab->startx = i;
            lab->starty = j;
            start = true;
         }
         if(lab->lab[i][j] == 'X')
         {
            if(treasure)
            {
               fprintf(stderr, "Error: Multiple treasure positions\n");
               exit(EXIT_FAILURE);
            }
            lab->treasurex = i;
            lab->treasurey = j;
            treasure = true;
         }
      }
   }
   if(!start)
   {
      fprintf(stderr, "Error: No start position\n");
      exit(EXIT_FAILURE);
   }
   if(!treasure)
   {
      fprintf(stderr, "Error: No treasure position\n");
      exit(EXIT_FAILURE);
   }
}



/*-----------------------------------------------------------------------
//
// Function: LabRead()
//
//   Read a labyrinth from an input stream and return it.
//
// Global Variables: -
//
// Side Effects    : Input, memory.
//
/----------------------------------------------------------------------*/

Lab_p LabRead(FILE* in)
{
   Lab_p handle = xmalloc(sizeof(LabCell));
   char  *inp;

   handle->maxrow = 0;

   while((inp=fgets(handle->lab[handle->maxrow], MAXCOLS, in)))
   {
      handle->maxrow++;
      if(handle->maxrow>MAXROWS)
      {
         fprintf(stderr, "Labyrinth too long\n");
         exit(EXIT_FAILURE);
      }
   }
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: LabPrint()
//
//   Print the current state of a labyrinth.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void LabPrint(FILE* out, Lab_p lab)
{
   int i;

   for(i=0; i<lab->maxrow; i++)
   {
      fprintf(out, "%s", lab->lab[i]);
   }
}



/*-----------------------------------------------------------------------
//
// Function: LabPrintCost()
//
//   Print the current state of a labyrinth.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void LabPrintCost(FILE* out, Lab_p lab)
{
   int i,j;


   for(i=0; i<lab->maxrow; i++)
   {
      for(j=0; j<lab->maxcol; j++)
      {
         if(lab->costs[i][j]!=LONG_MAX)
         {
            fprintf(out, "|%3ld", lab->costs[i][j]);
         }
         else
         {
            fprintf(out, "|###");
         }
      }
      fprintf(out, "|\n");
   }
}


/*-----------------------------------------------------------------------
//
// Function: LocalCost()
//
//   Return the cost of entering a field. In the basic case this is 1,
//   but for special values '0' - '9' it is the numerical value of the
//   digit.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long LocalCost(char field)
{
   if(field >= '0' && field <= '9')
   {
      return field - '0';
   }
   return 1;
}



/*-----------------------------------------------------------------------
//
// Function: LabSolve()
//
//   For all reachable fields, compute the cost of going there.
//
// Global Variables: -
//
// Side Effects    : Computes the lab costs and directions.
//
/----------------------------------------------------------------------*/

void LabSolve(Lab_p lab, long cost, RelPos moves[], int x, int y)
{
   int newx, newy;
   int i;
   long tmpcost;

   tmpcost = cost+LocalCost(lab->lab[x][y]);

   if(lab->lab[x][y] == '#')
   {
      return;
   }
   if(tmpcost >= lab->costs[x][y])
   {
      /* There already is a shorter way to here */
      return;
   }
   lab->costs[x][y] = tmpcost;
   if(lab->lab[x][y] == 'X')
   {
      /* No way leaving the goal is interesting */
      return;
   }
   for(i=0; moves[i].dx!=INT_MAX; i++)
   {
      newx = x+moves[i].dx;
      newy = y+moves[i].dy;
      LabSolve(lab, lab->costs[x][y], moves, newx,newy);
   }
}


/*-----------------------------------------------------------------------
//
// Function: LabMarkWay()
//
//   After LabSolve() has solved the lab, mark the best way.
//
// Global Variables: -
//
// Side Effects    : Marks the lab.
//
/----------------------------------------------------------------------*/

void LabMarkWay(Lab_p lab, RelPos moves[])
{
   int x = lab->treasurex, y = lab->treasurey, newx, newy;
   long cost, bestcost;
   int i, bestdir=0;

   while(lab->lab[x][y] != 'S')
   {
      if(lab->lab[x][y] != 'X')
      {
         lab->lab[x][y] = '.';
      }

      /* Find the cheapest neighbour */
      bestcost = LONG_MAX;
      for(i=0; moves[i].dx!=INT_MAX; i++)
      {
         newx = x+moves[i].dx;
         newy = y+moves[i].dy;

         if((lab->lab[newx][newy]!='#') &&
            ((cost = lab->costs[newx][newy]) < bestcost))
         {
            bestdir = i;
            bestcost = cost;
         }
      }
      x += moves[bestdir].dx;
      y += moves[bestdir].dy;
   }
}

/*-----------------------------------------------------------------------
//
// Function:
//
//
//
// Global Variables:
//
// Side Effects    :
//
/----------------------------------------------------------------------*/

void LabPrintDot(FILE* out, Lab_p lab, RelPos moves[])
{
   int x,y,i,newx,newy;
   char c;

   fprintf(out, "digraph laby {\n   rankdir = LR\n"
           "   node [fontsize=18]\n   edge [fontsize=14, dir=none]\n\n");
   for(y=0; y<lab->maxrow; y++)
   {
      /* First print all nodes */
      for(x=0; x<lab->maxcol; x++)
      {
         c = lab->lab[x][y];
         assert(c!='\n');
         switch(c)
         {
         case '#':
               break;
         case 'X':
         case 'S':
               fprintf(out, "   n%03d%03d [label=\"%c\"]\n", x,y,c);
               break;
         default:
               fprintf(out, "   n%03d%03d\n", x,y);
         }
      }
   }
   /* Now print all links. Because the graph is symmetric, we only
    * print links that go up and right*/
   for(y=0; y<lab->maxrow; y++)
   {
      /* First print all nodes */
      for(x=0; x<lab->maxcol; x++)
      {
         c = lab->lab[x][y];
         if(c!='#')
         {
            for(i=0; moves[i].dx!=INT_MAX; i++)
            {
               if((moves[i].dx > 0) ||
                  (moves[i].dx = 0 && moves[i].dy >0))
               {
                  newx = x+moves[i].dx;
                  newy = y+moves[i].dy;
                  if(lab->lab[newx][newy]!='#')
                  {
                     fprintf(stdout, "n%03d%03d->n%03d%03d\n",
                             x,y,newx,newy);
                  }
               }
            }
         }
      }

   }
   fprintf(out, "}\n");
}


/*-----------------------------------------------------------------------
//
// Function: main()
//
//
//
// Global Variables:
//
// Side Effects    :
//
/----------------------------------------------------------------------*/

int main(int argc, char *argv[])
{
   FILE *in = stdin;
   Lab_p lab;
   long cost;
   int arg, args=argc,  farg;
   RelPos *moves = straight_moves;
   bool print_costs = false;
   bool print_graph = false;

   for(arg = 1; arg<argc; arg++)
   {
      if(strcmp(argv[arg], "-a")==0)
      {
         moves = all_moves;
         args--;
      }
      else if(strcmp(argv[arg], "-c")==0)
      {
         print_costs = true;
         args--;
      }
      else if(strcmp(argv[arg], "-g")==0)
      {
         print_graph = true;
         args--;
      }
      else
      {
         farg = arg;
      }
   }

   if(args > 2)
   {
      fprintf(stderr, "Usage: %s [-a] [<file>]\n", argv[0]);
      exit(EXIT_FAILURE);
   }
   if(args == 2)
   {
      in = fopen(argv[farg], "r");
      if(!in)
      {
         perror(argv[0]);
         exit(EXIT_FAILURE);
      }
   }
   lab = LabRead(in);
   LabInit(lab);
   LabSolve(lab, 0, moves,lab->startx, lab->starty);
   if(print_costs)
   {
      LabPrintCost(stdout, lab);
   }
   if(print_graph)
   {
      LabPrintDot(stdout, lab, moves);
   }
   if((cost = lab->costs[lab->treasurex][lab->treasurey]) == LONG_MAX)
   {
      LabPrint(stdout, lab);
      fprintf(stdout, "There is no way from S to X\n");
   }
   else
   {
      LabMarkWay(lab, moves);
      LabPrint(stdout, lab);
      printf("Cost: %ld\n", cost);
   }
   exit(EXIT_SUCCESS);
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/
